/*
 * Password Management Servlets (PWM)
 * http://www.pwm-project.org
 *
 * Copyright (c) 2006-2009 Novell, Inc.
 * Copyright (c) 2009-2018 The PWM Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package password.pwm.config.value;

import com.google.gson.reflect.TypeToken;
import password.pwm.config.PwmSetting;
import password.pwm.config.PwmSettingSyntax;
import password.pwm.config.StoredValue;
import password.pwm.config.value.data.FormConfiguration;
import password.pwm.error.PwmOperationalException;
import password.pwm.util.java.JavaHelper;
import password.pwm.util.java.JsonUtil;
import password.pwm.util.java.StringUtil;
import password.pwm.util.java.XmlElement;
import password.pwm.util.java.XmlFactory;
import password.pwm.util.secure.PwmSecurityKey;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

public class FormValue extends AbstractValue implements StoredValue
{
    final List<FormConfiguration> values;

    private boolean needsXmlUpdate;

    public FormValue( final List<FormConfiguration> values )
    {
        this.values = values;
    }

    public static StoredValueFactory factory( )
    {
        return new StoredValueFactory()
        {
            public FormValue fromJson( final String input )
            {
                if ( input == null )
                {
                    return new FormValue( Collections.<FormConfiguration>emptyList() );
                }
                else
                {
                    List<FormConfiguration> srcList = JsonUtil.deserialize( input, new TypeToken<List<FormConfiguration>>()
                    {
                    } );
                    srcList = srcList == null ? Collections.<FormConfiguration>emptyList() : srcList;
                    while ( srcList.contains( null ) )
                    {
                        srcList.remove( null );
                    }
                    return new FormValue( Collections.unmodifiableList( srcList ) );
                }
            }

            public FormValue fromXmlElement( final PwmSetting pwmSetting, final XmlElement settingElement, final PwmSecurityKey key )
                    throws PwmOperationalException
            {
                final boolean oldType = PwmSettingSyntax.LOCALIZED_STRING_ARRAY.toString().equals(
                        settingElement.getAttributeValue( "syntax" ) );
                final List<XmlElement> valueElements = settingElement.getChildren( "value" );
                final List<FormConfiguration> values = new ArrayList<>();
                for ( final XmlElement loopValueElement  : valueElements )
                {
                    final String value = loopValueElement.getText();
                    if ( value != null && value.length() > 0 && loopValueElement.getAttributeValue( "locale" ) == null )
                    {
                        if ( oldType )
                        {
                            values.add( FormConfiguration.parseOldConfigString( value ) );
                        }
                        else
                        {
                            values.add( JsonUtil.deserialize( value, FormConfiguration.class ) );
                        }
                    }
                }
                final FormValue formValue = new FormValue( values );
                formValue.needsXmlUpdate = oldType;
                return formValue;
            }
        };
    }

    public List<XmlElement> toXmlValues( final String valueElementName, final PwmSecurityKey pwmSecurityKey  )
    {
        final List<XmlElement> returnList = new ArrayList<>();
        for ( final FormConfiguration value : values )
        {
            final XmlElement valueElement = XmlFactory.getFactory().newElement( valueElementName );
            valueElement.addText( JsonUtil.serialize( value ) );
            returnList.add( valueElement );
        }
        return returnList;
    }

    public List<FormConfiguration> toNativeObject( )
    {
        return Collections.unmodifiableList( values );
    }

    public List<String> validateValue( final PwmSetting pwmSetting )
    {
        if ( pwmSetting.isRequired() )
        {
            if ( values == null || values.size() < 1 || values.get( 0 ) == null )
            {
                return Collections.singletonList( "required value missing" );
            }
        }

        final Set<String> seenNames = new HashSet<>();
        for ( final FormConfiguration loopConfig : values )
        {
            if ( seenNames.contains( loopConfig.getName().toLowerCase() ) )
            {
                return Collections.singletonList( "each form name must be unique: " + loopConfig.getName() );
            }
            seenNames.add( loopConfig.getName().toLowerCase() );
        }

        for ( final FormConfiguration loopConfig : values )
        {
            try
            {
                loopConfig.validate();
            }
            catch ( PwmOperationalException e )
            {
                return Collections.singletonList( "format error: " + e.getErrorInformation().toDebugStr() );
            }
        }

        return Collections.emptyList();
    }

    public boolean isNeedsXmlUpdate( )
    {
        return needsXmlUpdate;
    }

    public String toDebugString( final Locale locale )
    {
        if ( values != null && !values.isEmpty() )
        {
            final StringBuilder sb = new StringBuilder();
            for ( final FormConfiguration formRow : values )
            {
                sb.append( "FormItem Name:" ).append( formRow.getName() ).append( "\n" );
                sb.append( " Type:" ).append( formRow.getType() );
                sb.append( " Min:" ).append( formRow.getMinimumLength() );
                sb.append( " Max:" ).append( formRow.getMaximumLength() );
                sb.append( " ReadOnly:" ).append( formRow.isReadonly() );
                sb.append( " Required:" ).append( formRow.isRequired() );
                sb.append( " Confirm:" ).append( formRow.isConfirmationRequired() );
                sb.append( " Unique:" ).append( formRow.isUnique() );
                sb.append( " Multi-Value:" ).append( formRow.isMultivalue() );
                sb.append( " Source:" ).append( formRow.getSource() );
                sb.append( "\n" );
                sb.append( " Label:" ).append( JsonUtil.serializeMap( formRow.getLabels() ) ).append( "\n" );
                sb.append( " Description:" ).append( JsonUtil.serializeMap( formRow.getDescription() ) ).append( "\n" );
                if ( formRow.getType() == FormConfiguration.Type.select && JavaHelper.isEmpty( formRow.getSelectOptions() ) )
                {
                    sb.append( " Select Options: " ).append( JsonUtil.serializeMap( formRow.getSelectOptions() ) ).append( "\n" );
                }
                if ( !StringUtil.isEmpty( formRow.getRegex() ) )
                {
                    sb.append( " Regex:" ).append( formRow.getRegex() )
                            .append( " Regex Error:" ).append( JsonUtil.serializeMap( formRow.getRegexErrors() ) )
                            .append( "\n" );
                }
                if ( formRow.getType() == FormConfiguration.Type.photo )
                {
                    sb.append( " MimeTypes: " ).append( StringUtil.collectionToString( formRow.getMimeTypes() ) ).append( "\n" );
                    sb.append( " MaxSize: " ).append( formRow.getMaximumSize() ).append( "\n" );
                }

            }
            return sb.toString();
        }
        else
        {
            return "";
        }
    }

}
