/*
 * Password Management Servlets (PWM)
 * http://www.pwm-project.org
 *
 * Copyright (c) 2006-2009 Novell, Inc.
 * Copyright (c) 2009-2018 The PWM Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package password.pwm.config.value;

import com.google.gson.reflect.TypeToken;
import password.pwm.config.PwmSetting;
import password.pwm.config.StoredValue;
import password.pwm.error.PwmOperationalException;
import password.pwm.util.java.JsonUtil;
import password.pwm.util.java.XmlElement;
import password.pwm.util.java.XmlFactory;
import password.pwm.util.secure.PwmSecurityKey;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.TreeSet;

public class OptionListValue extends AbstractValue implements StoredValue
{
    final Set<String> values;

    public OptionListValue( final Set<String> values )
    {
        this.values = new TreeSet( values );
    }

    public static StoredValueFactory factory( )
    {
        return new StoredValueFactory()
        {
            public OptionListValue fromJson( final String input )
            {
                if ( input == null )
                {
                    return new OptionListValue( Collections.emptySet() );
                }
                else
                {
                    Set<String> srcList = JsonUtil.deserialize( input, new TypeToken<Set<String>>()
                    {
                    } );
                    srcList = srcList == null ? Collections.emptySet() : srcList;
                    while ( srcList.contains( null ) )
                    {
                        srcList.remove( null );
                    }
                    return new OptionListValue( Collections.unmodifiableSet( srcList ) );
                }
            }

            public OptionListValue fromXmlElement( final PwmSetting pwmSetting, final XmlElement settingElement, final PwmSecurityKey key )
                    throws PwmOperationalException
            {
                final List<XmlElement> valueElements = settingElement.getChildren( "value" );
                final Set<String> values = new TreeSet<>();
                for ( final XmlElement loopValueElement : valueElements )
                {
                    final String value = loopValueElement.getText();
                    if ( value != null && !value.trim().isEmpty() )
                    {
                        values.add( value );
                    }
                }
                return new OptionListValue( values );
            }
        };
    }

    public List<XmlElement> toXmlValues( final String valueElementName, final PwmSecurityKey pwmSecurityKey  )
    {
        final List<XmlElement> returnList = new ArrayList<>();
        for ( final String value : values )
        {
            final XmlElement valueElement = XmlFactory.getFactory().newElement( valueElementName );
            valueElement.addText( value );
            returnList.add( valueElement );
        }
        return returnList;
    }

    public Set<String> toNativeObject( )
    {
        return Collections.unmodifiableSet( values );
    }

    public List<String> validateValue( final PwmSetting pwmSetting )
    {
        return Collections.emptyList();
    }

    public String toDebugString( final Locale locale )
    {
        final StringBuilder sb = new StringBuilder();
        for ( final Iterator valueIterator = values.iterator(); valueIterator.hasNext(); )
        {
            sb.append( valueIterator.next() );
            if ( valueIterator.hasNext() )
            {
                sb.append( "\n" );
            }
        }
        return sb.toString();
    }
}
